@@ -1,6 +1,6 @@
 /*
     mpg321 - a fully free clone of mpg123.
-    Copyright (C) 2001 Joe Drew
+    ao.c: Copyright (C) 2001, 2002 Joe Drew
     
     Originally based heavily upon:
     plaympeg - Sample MPEG player using the SMPEG library
@@ -29,6 +29,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <signal.h>
 
 void set_play_device(char *devicename)
 {
@@ -57,13 +58,15 @@ void set_play_device(char *devicename)
         options.opt |= MPG321_USE_ARTS;
     }
 
+    else if (strcmp(devicename, "alsa09") == 0)
+    {
+        options.opt |= MPG321_USE_ALSA09;
+    }
+    
     else
     {
-        fprintf(stderr, "Unrecognised/unsupported default libao output device %s.\n"
-            "Please e-mail drew@debian.org with this error message.\n"
-            "You can also edit your /etc/libao.conf or ~/.libao file(s).\n", 
-                devicename);
-        exit(1);
+        options.opt |= MPG321_USE_USERDEF;
+        options.devicetype = strdup(devicename);
     }
 }
 
@@ -91,11 +94,9 @@ void check_ao_default_play_device()
 void check_default_play_device()
 {
     /* check that no output devices are currently selected */
-    if (!(options.opt & MPG321_USE_OSS || options.opt & MPG321_USE_STDOUT
-        || options.opt & MPG321_USE_ALSA || options.opt & MPG321_USE_ESD 
-        || options.opt & MPG321_USE_NULL || options.opt & MPG321_USE_WAV
-        || options.opt & MPG321_USE_ARTS || options.opt & MPG321_USE_AU
-        || options.opt & MPG321_USE_CDR))
+    if (!(options.opt & (MPG321_USE_OSS | MPG321_USE_STDOUT | MPG321_USE_ALSA | MPG321_USE_ESD 
+                         | MPG321_USE_NULL | MPG321_USE_WAV | MPG321_USE_ARTS | MPG321_USE_AU 
+                         | MPG321_USE_CDR | MPG321_USE_ALSA09 | MPG321_USE_USERDEF)))
     {
         ao_info *default_info;
         
@@ -119,7 +120,7 @@ void check_default_play_device()
 #endif
 
 #ifdef HAS_PUTENV
-	    putenv("ESD_NO_SPAWN=1");
+            putenv("ESD_NO_SPAWN=1");
 #else
 #ifdef HAS_SETENV
             setenv("ESD_NO_SPAWN", "1", 0);
@@ -128,7 +129,7 @@ void check_default_play_device()
             driver_id = ao_default_driver_id();
 
 #ifdef HAS_PUTENV
-	    if (unset) putenv("ESD_NO_SPAWN");
+            if (unset) putenv("ESD_NO_SPAWN");
 #else
 #ifdef HAS_UNSETENV
             if (unset) unsetenv("ESD_NO_SPAWN");
@@ -206,6 +207,16 @@ int playdevice_is_live()
         {
             driver_id = ao_driver_id("sun");
         }
+        
+        else if (options.opt & MPG321_USE_ALSA09)
+        {
+            driver_id = ao_driver_id("alsa09");
+        }
+        
+        else if (options.opt & MPG321_USE_USERDEF)
+        {
+            driver_id = ao_driver_id(options.devicetype);
+        }
 
     return (ao_driver_info(driver_id)->type == AO_TYPE_LIVE);
 }
@@ -214,6 +225,10 @@ void open_ao_playdevice(struct mad_header const *header)
 {
         ao_sample_format format;
 
+        /* Because these can sometimes block, we stop our custom signal handler,
+           and restore it afterwards */
+        signal(SIGINT, SIG_DFL);
+        
         format.bits = 16;
         format.rate = header->samplerate;
         format.channels = (options.opt & MPG321_FORCE_STEREO) ? 2 : MAD_NCHANNELS(header);
@@ -301,39 +316,66 @@ void open_ao_playdevice(struct mad_header const *header)
             }
         }
         
-        else if(options.opt & MPG321_USE_ESD)
+        else if (options.opt & MPG321_USE_USERDEF)
         {
             ao_option *ao_options = NULL;
-            int driver_id = ao_driver_id("esd");
-           
-            if(options.device)
-                ao_append_option(&ao_options, "host", options.device);
-        
-            if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
+            int driver_id = ao_driver_id(options.devicetype);
+            
+            if (driver_id < 0)
             {
-                fprintf(stderr, "Error opening libao esd driver. (Is ESD running?)\n");
+                fprintf(stderr, "Can't open unknown ao driver %s\n", options.devicetype);
                 exit(1);
             }
-        }
-        
-        else if(options.opt & MPG321_USE_ARTS)
-        {
-            ao_option *ao_options = NULL;
-            int driver_id = ao_driver_id("arts");
-        
-            if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
+
+            if (playdevice_is_live())
             {
-                fprintf(stderr, "Error opening libao arts driver. (Is aRts running?)\n");
-                exit(1);
+                if (options.device)
+                {
+                    fprintf(stderr, "Can't set output device to %s for unknown ao plugin %s",
+                            options.device, options.devicetype);
+                }
+
+                if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
+                {
+                    fprintf(stderr, "Error opening unknown libao %s driver. (Is device in use?)\n",
+                        options.devicetype);
+                    exit(1);
+                }
             }
-        }
-        
-        else if(options.opt & MPG321_USE_ALSA)
+
+            else
+            {
+                if (options.device)
+                {
+                    /* Just assume that options.device is a filename. The user can shoot 
+                       themselves in the foot all they like... */
+                    if((playdevice=ao_open_file(driver_id, options.device, 1 /*overwrite*/,
+                            &format, ao_options))==NULL)
+                    {
+                        fprintf(stderr, "Error opening unknown libao %s file driver for file %s. (Do you have write permissions?)\n",
+                                options.devicetype, options.device);
+                        exit(1);
+                    }
+                }
+                
+                else
+                {
+                    fprintf(stderr, "Filename must be specified (with -a filename) for unknown ao driver %s\n",
+                            options.devicetype);
+                }
+            }
+        } else {
+            /* Hack-tacular. This code tries to the device as specified; if it can't, it'll
+               fall through to the other devices, trying each in turn. If the user specified
+               a device to use, though, it won't fall through: principle of least surprise */
+        int opened = 0;
+
+        if(!opened && options.opt & MPG321_USE_ALSA)
         {
             ao_option *ao_options = NULL;
             int driver_id = ao_driver_id("alsa");
             char *c;
-           
+
             if (options.device)
             {
                 if ((c = strchr(options.device, ':')) == NULL || strlen(c+1) < 1)
@@ -341,53 +383,124 @@ void open_ao_playdevice(struct mad_header const *header)
                     fprintf(stderr, "Poorly formed ALSA card:device specification %s", options.device);
                     exit(1);
                 }
-                
+
                 *(c++) = '\0'; /* change the : to a null to create two separate strings */
-                
+
                 ao_append_option(&ao_options, "card", options.device);
                 ao_append_option(&ao_options, "dev", c);
             }
-        
+
             if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
             {
-                fprintf(stderr, "Error opening libao alsa driver. (Is device in use?)\n");
-                exit(1);
-            }
+                if (options.device)
+                {
+                    fprintf(stderr, "Can't open libao driver with device %s (is device in use?)\n",
+                        options.device);
+                    exit(1);
+                }
+                else
+                options.opt |= MPG321_USE_ALSA09;
+            } else opened++;
         }
-        
-        else if(options.opt & MPG321_USE_OSS)
+
+        if(!opened && options.opt & MPG321_USE_ALSA09)
+        {
+            ao_option *ao_options = NULL;
+            int driver_id = ao_driver_id("alsa09");
+
+            if(options.device)
+                ao_append_option(&ao_options, "dev", options.device);
+
+            if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
+            {
+                if (options.device)
+                {
+                    fprintf(stderr, "Can't open libao driver with device %s (is device in use?)\n",
+                        options.device);
+                    exit(1);
+                }
+                else
+                options.opt |= MPG321_USE_OSS;
+            } else opened++;
+        }
+
+        if(!opened &&  options.opt & MPG321_USE_OSS)
         {
             ao_option *ao_options = NULL;
             int driver_id = ao_driver_id("oss");
-           
+
             if(options.device)
                 ao_append_option(&ao_options, "dsp", options.device);
-        
+
             if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
             {
-                fprintf(stderr, "Error opening libao oss driver. (Is device in use?)\n");
-                exit(1);
-            }
+                if (options.device)
+                {
+                    fprintf(stderr, "Can't open libao driver with device %s (is device in use?)\n",
+                        options.device);
+                    exit(1);
+                }
+                else
+                options.opt |= MPG321_USE_SUN;
+            } else opened++;
         }
 
-        else if(options.opt & MPG321_USE_SUN)
+        if(!opened && options.opt & MPG321_USE_SUN)
         {
             ao_option *ao_options = NULL;
             int driver_id = ao_driver_id("sun");
-           
+
             if(options.device)
                 ao_append_option(&ao_options, "dev", options.device);
+
+            if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
+            {
+                if (options.device)
+                {
+                    fprintf(stderr, "Can't open libao driver with device %s (is device in use?)\n",
+                        options.device);
+                    exit(1);
+                }
+                else
+                options.opt |= MPG321_USE_ESD;
+            }
+        }
+
+        if(!opened && options.opt & MPG321_USE_ESD)
+        {
+            ao_option *ao_options = NULL;
+            int driver_id = ao_driver_id("esd");
+
+            if(options.device)
+                ao_append_option(&ao_options, "host", options.device);
+
+            if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
+            {
+                if (options.device)
+                {
+                    fprintf(stderr, "Can't open libao driver with device %s (is device in use?)\n",
+                        options.device);
+                    exit(1);
+                }
+                else
+                options.opt |= MPG321_USE_ARTS;
+            } else opened++;
+        }
+
+        if(!opened && options.opt & MPG321_USE_ARTS)
+        {
+            ao_option *ao_options = NULL;
+            int driver_id = ao_driver_id("arts");
         
             if((playdevice=ao_open_live(driver_id, &format, ao_options))==NULL)
             {
-                fprintf(stderr, "Error opening libao sun driver. (Is device in use?)\n");
+                fprintf(stderr, "Can't find a suitable libao driver. (Is device in use?)\n");
                 exit(1);
             }
         }
         
-        else /* how did we get here? see ao.c: set_default_play_device()*/
-        {
-            fprintf(stderr, "Internal error: mpg321 has no default device.\n"
-                    "Please e-mail drew@debian.org with specifics if this happens.\n");
-        }
+    }
+
+    /* Restore signal handler */
+    signal(SIGINT, handle_signals);
 }
