@@ -61,12 +61,6 @@ enum mad_flow read_from_mmap(void *data, struct mad_stream *stream)
         exit(1);
     }
     
-    if(!playbuf->buf)
-    {
-        fprintf(stderr, "read_from_mmap called with null buf!\n");
-        exit(1);
-    }
-
     mpegdata = playbuf->buf;
 
     if(status == MPG321_REWINDING)
@@ -90,6 +84,7 @@ enum mad_flow read_from_mmap(void *data, struct mad_stream *stream)
 enum mad_flow read_from_fd(void *data, struct mad_stream *stream)
 {
     buffer *playbuf = data;
+    int bytes_to_preserve = stream->bufend - stream->next_frame;
     
     if(playbuf->done)
     {
@@ -108,7 +103,12 @@ enum mad_flow read_from_fd(void *data, struct mad_stream *stream)
         fprintf(stderr, "read_from_fd called with null buf!\n");
     }
     
-    if((playbuf->length = read(playbuf->fd, playbuf->buf, BUF_SIZE))==-1)
+    /* need to preserve the bytes which comprise an incomplete
+       frame, that mad has passed back to us */
+    if (bytes_to_preserve)
+        memmove(playbuf->buf, stream->next_frame, bytes_to_preserve);
+
+    if( !(read(playbuf->fd, playbuf->buf + bytes_to_preserve, BUF_SIZE - bytes_to_preserve) > 0) )
         playbuf->done = 1;
 
     mad_stream_buffer(stream, playbuf->buf, playbuf->length);
@@ -167,13 +167,9 @@ char * modestringucase(enum mad_mode mode)
 
 enum mad_flow read_header(void *data, struct mad_header const * header)
 {
-    static char current_file[PATH_MAX] = "";
-
     char long_currenttime_str[14]; /* this *will* fill if you're using 100000+ minute mp3s */
     char long_remaintime_str[14];
     
-    int new_file = 0;
-
     buffer *playbuf = (buffer *)data;
     mad_timer_t time_remaining;
     
@@ -192,35 +188,29 @@ enum mad_flow read_header(void *data, struct mad_header const * header)
             return mf;
     }
 
-    if(strncmp(current_file, playbuf->filename, PATH_MAX) != 0)
+    /* Stop playing if -n is used, and we're at the frame specified. */
+    if ((playbuf->max_frames != -1) && (current_frame > playbuf->max_frames))
     {
-        new_file = 1;
-        current_frame = 0;
-        strncpy(current_file, playbuf->filename, PATH_MAX);
-        current_file[PATH_MAX-1] = '\0';
+      status = MPG321_STOPPED;
+      return MAD_FLOW_STOP;
     }
 
-#if 0
-    /* I don't know if I agree with this, but it seems to be
-       what mpg123 does. I will revert this if anybody complains
-       at all. */    
-    if (current_frame > playbuf->num_frames)
-    {
-        status = MPG321_STOPPED;
-        return MAD_FLOW_STOP;
-    }
-#endif
-
     current_frame++;
 
     mad_timer_add(&current_time, header->duration);
 
-    if(options.opt & MPG321_VERBOSE_PLAY || options.opt & MPG321_REMOTE_PLAY)
+    if(options.opt & (MPG321_VERBOSE_PLAY | MPG321_REMOTE_PLAY))
     {
         mad_timer_string(current_time, long_currenttime_str, "%.2u:%.2u.%.2u", MAD_UNITS_MINUTES,
                             MAD_UNITS_CENTISECONDS, 0);
+
+        if (mad_timer_compare(playbuf->duration, mad_timer_zero) == 0)
+            time_remaining = current_time;
+        else
+            time_remaining = playbuf->duration;
+
         mad_timer_negate(&current_time);
-        time_remaining = playbuf->duration;
+
         mad_timer_add(&time_remaining, current_time);
         mad_timer_negate(&current_time);
 
@@ -236,15 +226,9 @@ enum mad_flow read_header(void *data, struct mad_header const * header)
         playbuf->times[current_frame] = current_time;
     }
     
-    else
-    {
-        /* We've exceeded our estimate. Printing of any data we have will output crap,
-           so we just return here to make sure we don't cut off the mp3. */
-        return MAD_FLOW_CONTINUE;
-    }    
-
-    if (new_file)
+    if (file_change)
     {
+        file_change = 0;
         if (options.opt & MPG321_REMOTE_PLAY)
         {
             printf("@S 1.0 %d %d %s %d %ld %d %d %d %d %ld %d\n", header->layer, header->samplerate,
@@ -293,7 +277,7 @@ enum mad_flow read_header(void *data, struct mad_header const * header)
         if (!options.skip_printing_frames 
             || (options.skip_printing_frames && !(current_frame % options.skip_printing_frames)))
             fprintf(stderr, "Frame# %5lu [%5lu], Time: %s [%s], \r", current_frame, 
-                    playbuf->num_frames - current_frame, long_currenttime_str, long_remaintime_str);
+                    playbuf->num_frames > 0 ? playbuf->num_frames - current_frame : 0, long_currenttime_str, long_remaintime_str);
     }
     
     else if (options.opt & MPG321_REMOTE_PLAY)
@@ -509,7 +493,8 @@ void pause_play(buffer *buf, playlist *pl)
     if (strlen(file) == 0)
     {
         status = MPG321_PAUSED;
-        strcpy(file, buf->filename);
+        strncpy(file, buf->filename, PATH_MAX);
+        file[PATH_MAX-1]='\0';
         clear_remote_file(pl);
         seek = current_frame;
         printf("@P 1\n");
@@ -603,14 +588,13 @@ int calc_length(char *file, buffer *buf)
 
     if (lseek(f, -128, SEEK_END) < 0)
     {
-        mpg321_error(file);
+        /* File must be very short or empty. Forget it. */
         close(f);
         return -1;
     }    
 
     if (read(f, buffer, 3) != 3)
     {
-        fprintf(stderr, "Couldnt read 3 bytes from %s\n", file);
         close(f);
         return -1;
     }
