@@ -33,6 +33,7 @@
 #include <signal.h>
 #include <limits.h>
 #include <sys/time.h>
+#include <assert.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -43,14 +44,19 @@
 #include "getopt.h" /* GNU getopt is needed, so I included it */
 #include "mpg321.h"
 
+#include <id3tag.h>
+
 int shuffle_play;
 int stop_playing_file = 0;
 int quit_now = 0;
 char *playlist_file;
 ao_device *playdevice=NULL;
 mad_timer_t current_time;
-mpg321_options options = { 0, NULL, 0 , 0, 0};
+mpg321_options options = { 0, NULL, NULL, 0 , 0, 0};
 int status = MPG321_STOPPED;
+int file_change = 0;
+
+char *id3_get_tag (struct id3_tag const *tag, char const *what, unsigned int maxlen);
 
 void mpg123_boilerplate()
 {
@@ -78,7 +84,8 @@ void usage(char *argv0)
         "   --gain N or -g N         Set gain (audio volume) to N (0-100)\n"
         "   --skip N or -k N         Skip N frames into the file\n"
         "   --verbose or -v          Be more verbose in playing files\n"
-        "   -o dt                    Set output devicetype to dt [esd,alsa,arts,sun,oss]\n"
+        "   -o dt                    Set output devicetype to dt\n" 
+    "                                [esd,alsa(09),arts,sun,oss]\n"
         "   --audiodevice N or -a N  Use N for audio-out\n"
         "   --stdout or -s           Use stdout for audio-out\n"
         "   --au N                   Use au file N for output\n"
@@ -94,7 +101,7 @@ void usage(char *argv0)
         "   --version or -V          Print version information\n"
         "\n"
         "This version of mpg321 has been configured with " AUDIO_DEFAULT " as its default\n"
-        "output device.\n" , argv0);
+        "libao output device.\n" , argv0);
 }
 
 /* retsigtype is defined by configure;
@@ -106,10 +113,10 @@ RETSIGTYPE handle_signals(int sig)
     
     gettimeofday(&curr_tv, NULL);
     
-    if (sig != -1)
+    if (sig == SIGINT)
     {
         if (((curr_tv.tv_sec - last_signal.tv_sec) * 1000000 +  
-            (curr_tv.tv_usec - last_signal.tv_usec)) < 1000000) /* Allow a short delay to still
+            (curr_tv.tv_usec - last_signal.tv_usec)) < 1000000) /* Allow a short delay to
                                                                   kill mpg321 */
         {
             quit_now = 1;
@@ -120,18 +127,99 @@ RETSIGTYPE handle_signals(int sig)
         
         stop_playing_file = 1;
    }
-   else /* initialize */
+   else if (sig == -1) /* initialize */
    {
         last_signal.tv_sec = curr_tv.tv_sec;
         last_signal.tv_usec = curr_tv.tv_usec;
    }
 }
 
+/*
+    Only shows ID3 tags if at least one of
+    TITLE, ARTIST, ALBUM, YEAR, COMMENT, GENRE
+    is present.
+*/
+static int show_id3(struct id3_tag const *tag)
+{
+    unsigned int i;
+    int print = 0;
+    char emptystring[31];
+    char *names[6];
+    struct {
+        int index;
+        char const *id;
+        char const *name;
+    } const info[] = {
+        { 0,    ID3_FRAME_TITLE,  "Title  : "   },
+        { 1,    ID3_FRAME_ARTIST, "  Artist: "  },
+        { 2,    ID3_FRAME_ALBUM,  "Album  : "   },
+        { 3,    ID3_FRAME_YEAR,   "  Year  : "  },
+        { 4,    ID3_FRAME_COMMENT,"Comment: "   },
+        { 5,    ID3_FRAME_GENRE,  "  Genre : "  }
+    };
+
+    memset (emptystring, ' ', 30);
+    emptystring[30] = '\0';
+    /*  Get ID3 tag if available, 30 chars except 4 for year  */
+    for (i=0; i<=5; i++)    {
+        names[i] = NULL;
+        names[i] = id3_get_tag(tag, info[i].id, (i==3) ? 4 : 30);
+    }
+    for (i=0; i<=5; i++)    {
+        if (names[i])   {
+            print = 1;
+            break;
+        }
+    }
+    if (!print) {
+        return 0;
+    }
+
+    if (options.opt & MPG321_REMOTE_PLAY)
+    {
+        printf("@I ID3:");
+
+        for (i=0; i<=5; i++)
+        {
+            if(!names[i])
+            {
+                printf(emptystring);
+            }
+            
+            else
+            {
+                printf(names[i]);
+                free(names[i]);
+            }
+        }
+        printf("\n");
+    }
+    
+    else
+    {
+        /* Emulate mpg123 original behaviour  */
+        for (i=0; i<=5; i++)    {
+            fprintf (stderr, "%s", info[i].name);
+            if (!names[i])  {
+                fprintf (stderr, emptystring);
+            }   else    {
+                fprintf (stderr, names[i]);
+                free (names[i]);
+            }
+            if (i%2) fprintf (stderr, "\n");
+        }
+    }
+
+    return 1;
+}
+
 int main(int argc, char *argv[])
 {
     int fd = 0;
     char *currentfile, old_dir[PATH_MAX];
     playlist *pl = NULL;
+    struct id3_file *id3struct = NULL;
+    struct id3_tag *id3tag = NULL;
 
     buffer playbuf;
     
@@ -149,6 +237,8 @@ int main(int argc, char *argv[])
 
     options.volume = MAD_F_ONE;
 
+    status = MPG321_PLAYING;
+    
     /* Get the command line options */
     parse_options(argc, argv, pl);
 
@@ -175,8 +265,6 @@ int main(int argc, char *argv[])
     {
         handle_signals(-1); /* initialize signal handler */
      
-        signal(SIGINT, handle_signals);
-
         remote_input_buf[0] = '\0';
     }
     
@@ -194,17 +282,95 @@ int main(int argc, char *argv[])
         if (quit_now) 
             break;
         
+        signal(SIGINT, SIG_DFL);
+        
         playbuf.buf = NULL;
         playbuf.fd = -1;
         playbuf.length = 0;
         playbuf.done = 0;
         playbuf.num_frames = 0;
-        strcpy(playbuf.filename,currentfile);
+        playbuf.max_frames = -1;
+        strncpy(playbuf.filename,currentfile, PATH_MAX);
+        playbuf.filename[PATH_MAX-1] = '\0';
+        
+        if (status == MPG321_PLAYING) 
+            file_change = 1;
 
         mad_timer_reset(&playbuf.duration);
         
         mad_timer_reset(&current_time);
 
+        if (!(options.opt & MPG321_QUIET_PLAY) && file_change)
+        {
+            id3struct = id3_file_open (currentfile, ID3_FILE_MODE_READONLY);
+
+            if (id3struct)
+            {
+                id3tag = id3_file_tag (id3struct);
+            
+                if (id3tag)
+                {
+                    show_id3 (id3tag);
+                }
+
+                id3_file_close (id3struct);
+            }
+        }
+
+        if (options.opt & MPG321_REMOTE_PLAY && file_change)
+        {
+            id3struct = id3_file_open (currentfile, ID3_FILE_MODE_READONLY);
+
+            if (id3struct)
+            {
+                id3tag = id3_file_tag (id3struct);
+            
+                if (id3tag)
+                {
+                    if (!show_id3(id3tag))
+                    {
+                        /* This shouldn't be necessary, but it appears that
+                           libid3tag doesn't necessarily know if there are no
+                           id3 tags on a given mp3 */
+                        char * basec = strdup(currentfile);
+                        char * basen = basename(basec);
+                        
+                        char * dot = strrchr(basen, '.');
+                        
+                        if (dot)
+                            *dot = '\0';
+                        
+                        printf("@I %s\n", basen);
+
+                        free(basec);
+                    }
+                }
+                
+                else
+                {
+                    fprintf(stderr, "Allocation error");
+                    exit(1);
+                }
+
+                id3_file_close (id3struct);
+            }
+            
+            else
+            {
+                char * basec = strdup(currentfile);
+                char * basen = basename(basec);
+                
+                char * dot = strrchr(basen, '.');
+                
+                if (dot)
+                    *dot = '\0';
+                
+                printf("@I %s\n", basen);
+
+                free(basec);
+            }
+        }
+
         /* Create the MPEG stream */
         /* Check if source is on the network */
         if((fd = raw_open(currentfile)) != 0 || (fd = http_open(currentfile)) != 0
@@ -212,6 +378,8 @@ int main(int argc, char *argv[])
         {
             playbuf.fd = fd;
             playbuf.buf = malloc(BUF_SIZE);
+            playbuf.length = BUF_SIZE;
+            
             mad_decoder_init(&decoder, &playbuf, read_from_fd, read_header, /*filter*/0,
                             output, /*error*/0, /* message */ 0);
         }
@@ -221,6 +389,8 @@ int main(int argc, char *argv[])
         {
             playbuf.fd = fileno(stdin);
             playbuf.buf = malloc(BUF_SIZE);
+            playbuf.length = BUF_SIZE;
+
             mad_decoder_init(&decoder, &playbuf, read_from_fd, read_header, /*filter*/0,
                             output, /*error*/0, /* message */ 0);
         }
@@ -246,14 +416,14 @@ int main(int argc, char *argv[])
             
             if (!S_ISREG(stat.st_mode))
             {
-                fprintf(stderr, "file %s is not a regular file!\n", currentfile);
                 continue;
             }
             
-            if (calc_length(currentfile, &playbuf) == -1)
-            {
-                fprintf(stderr, "Couldn't calculate duration of %s!\n",currentfile);
-                continue;
+            calc_length(currentfile, &playbuf);
+
+            if ((options.maxframes != -1) && (options.maxframes <= playbuf.num_frames))
+            { 
+                playbuf.max_frames = options.maxframes;
             }
             
             playbuf.frames = malloc((playbuf.num_frames + 1) * sizeof(void*));
@@ -306,26 +476,8 @@ int main(int argc, char *argv[])
             free(basec);
         }    
 
-        if (options.opt & MPG321_REMOTE_PLAY)
-        {
-            char * basec = strdup(currentfile);
-            char * basen = basename(basec);
-            
-            char * dot = strrchr(basen, '.');
-            
-            if (dot)
-                *dot = '\0';
-            
-            if (status == MPG321_PLAYING)
-            {
-                printf("@I %s\n", basen);
-                
-             //   strncpy(old_dir, basen, PATH_MAX);
-            }
-            
-            free(basec);
-        }
-        
+        signal(SIGINT, handle_signals);
+
         /* Every time the user gets us to rewind, we exit decoding,
            reinitialize it, and re-start it */
         while (1)
@@ -374,8 +526,8 @@ int main(int argc, char *argv[])
         else
         {
             free(playbuf.buf);
-            if (fd != fileno(stdin)) 
-                close(fd);
+            if (playbuf.fd != fileno(stdin)) 
+                close(playbuf.fd);
         }
     }
 
@@ -390,3 +542,93 @@ int main(int argc, char *argv[])
 
     return(0);
 }
+
+
+/* Convenience for retrieving already formatted id3 data
+ * what parameter is one of
+ *  ID3_FRAME_TITLE
+ *  ID3_FRAME_ARTIST
+ *  ID3_FRAME_ALBUM
+ *  ID3_FRAME_YEAR
+ *  ID3_FRAME_COMMENT
+ *  ID3_FRAME_GENRE
+ * It allocates a new string. Free it later.
+ * NULL if no tag or error.
+ */
+char *id3_get_tag (struct id3_tag const *tag, char const *what, unsigned int maxlen)
+{
+    struct id3_frame const *frame = NULL;
+    union id3_field const *field = NULL;
+    int nstrings;
+    int avail;
+    int j;
+    int tocopy;
+    int len;
+    char printable [1024];
+    char *retval = NULL;
+    id3_ucs4_t const *ucs4 = NULL;
+    id3_latin1_t *latin1 = NULL;
+
+    memset (printable, '\0', 1024);
+    avail = 1024;
+    if (strcmp (what, ID3_FRAME_COMMENT) == 0)
+    {
+        /*There may be sth wrong. I did not fully understand how to use
+            libid3tag for retrieving comments  */
+        j=0;
+        frame = id3_tag_findframe(tag, ID3_FRAME_COMMENT, j++);
+        if (!frame) return (NULL);
+        ucs4 = id3_field_getfullstring (&frame->fields[3]);
+        if (!ucs4) return (NULL);
+        latin1 = id3_ucs4_latin1duplicate (ucs4);
+        if (!latin1 || strlen(latin1) == 0) return (NULL);
+        len = strlen(latin1);
+        if (avail > len)
+            tocopy = len;
+        else
+            tocopy = 0;
+        if (!tocopy) return (NULL);
+        avail-=tocopy;
+        strncat (printable, latin1, tocopy);
+        free (latin1);
+    }
+    
+    else
+    {
+        frame = id3_tag_findframe (tag, what, 0);
+        if (!frame) return (NULL);
+        field = &frame->fields[1];
+        nstrings = id3_field_getnstrings(field);
+        for (j=0; j<nstrings; ++j)
+        {
+            ucs4 = id3_field_getstrings(field, j);
+            if (!ucs4) return (NULL);
+            if (strcmp (what, ID3_FRAME_GENRE) == 0)
+                ucs4 = id3_genre_name(ucs4);
+            latin1 = id3_ucs4_latin1duplicate(ucs4);
+            if (!latin1) break;
+            len = strlen(latin1);
+            if (avail > len)
+                tocopy = len;
+            else
+                tocopy = 0;
+            if (!tocopy) break;
+            avail-=tocopy;
+            strncat (printable, latin1, tocopy);
+            free (latin1);
+        }
+    }
+    retval = malloc (maxlen + 1);
+    if (!retval) return (NULL);
+
+    strncpy (retval, printable, maxlen);
+    retval[maxlen] = '\0';
+
+    len = strlen(printable);
+    if (maxlen > len)
+    {
+        memset (retval + len, ' ', maxlen - len);
+    }
+
+    return (retval);
+}
