@@ -39,7 +39,7 @@
 #include <mad.h>
 
 #define FAKEVERSION "0.59q"
-#define VERSIONDATE "2001/Nov/30"
+#define VERSIONDATE "2002/03/23"
 
 /* playlist structure */
 typedef struct pl
@@ -75,6 +75,9 @@ typedef struct
     /* total number of frames */
     unsigned long num_frames;
 
+    /* number of frames to play */
+    unsigned long max_frames;
+
     /* total duration of the file */
     mad_timer_t duration;
 
@@ -88,8 +91,10 @@ typedef struct
 typedef struct
 {
     int opt;
+    char *devicetype;
     char *device;
     signed long seek;
+    signed long maxframes;
     int volume;
     int skip_printing_frames;
 } mpg321_options;    
@@ -103,6 +108,7 @@ extern int shuffle_play;
 extern char *playlist_file;
 extern int quit_now;
 extern char remote_input_buf[PATH_MAX + 5];
+extern int file_change;
 
 extern int status;
 
@@ -117,25 +123,28 @@ enum
 
 enum
 {
-    MPG321_VERBOSE_PLAY  = 0x0001,
-    MPG321_QUIET_PLAY    = 0x0002,
-    MPG321_REMOTE_PLAY   = 0x0004,
+    MPG321_VERBOSE_PLAY  = 0x00000001,
+    MPG321_QUIET_PLAY    = 0x00000002,
+    MPG321_REMOTE_PLAY   = 0x00000004,
 
-    MPG321_USE_OSS       = 0x0010,
-    MPG321_USE_SUN       = 0x0020,
-    MPG321_USE_ALSA      = 0x0040,
-    MPG321_USE_ESD       = 0x0080,
-    MPG321_USE_ARTS      = 0x0100,
-    MPG321_USE_NULL      = 0x0200,
-    MPG321_USE_STDOUT    = 0x0400,
-    MPG321_USE_WAV       = 0x0800,
-    MPG321_USE_AU        = 0x1000,
-    MPG321_USE_CDR       = 0x2000,
-    MPG321_FORCE_STEREO  = 0x4000
+    MPG321_USE_OSS       = 0x00000010,
+    MPG321_USE_SUN       = 0x00000020,
+    MPG321_USE_ALSA      = 0x00000040,
+    MPG321_USE_ESD       = 0x00000080,
+    MPG321_USE_ARTS      = 0x00000100,
+    MPG321_USE_NULL      = 0x00000200,
+    MPG321_USE_STDOUT    = 0x00000400,
+    MPG321_USE_WAV       = 0x00000800,
+    MPG321_USE_AU        = 0x00001000,
+    MPG321_USE_CDR       = 0x00002000,
+    MPG321_USE_USERDEF   = 0x00004000,
+    MPG321_USE_ALSA09    = 0x00008000,
+    
+    MPG321_FORCE_STEREO  = 0x00010000
 };
 
 #define DEFAULT_PLAYLIST_SIZE 1024
-#define BUF_SIZE 4096 /* Size for read buffer for audio data */
+#define BUF_SIZE 1048576 /* Size for read buffer for audio data */
 
 /* playlist functions */
 playlist * new_playlist();
@@ -184,4 +193,6 @@ void parse_options(int argc, char *argv[], playlist *pl);
 void mpg321_error(char *file);
 void usage(char *);
 
+RETSIGTYPE handle_signals(int sig) ;
+
 #endif /* _MPG321_H_ */
